# -*- coding: utf-8 -*-
import click

from ..sierraclient import SierraClient, VERSION


@click.group(invoke_without_command=True)
@click.option('--url',
              default='https://hivdb.stanford.edu/graphql',
              help='URL of Sierra GraphQL Web Service.')
@click.option('--version', is_flag=True,
              help='Show client and the HIVDB algorithm version.')
@click.pass_context
def cli(ctx, url, version):
    """A Client of HIVDB Sierra GraphQL Web Service"""
    ctx.obj['URL'] = url
    client = ctx.obj['CLIENT'] = SierraClient(url)
    client.toggle_progress(True)
    if version:
        result = client.current_version()
        click.echo(
            'SierraPy {}; HIVdb {} ({})'
            .format(VERSION, result['text'], result['publishDate']))
        exit(0)
    elif not ctx.invoked_subcommand:
        click.echo(cli.get_help(ctx))
    """
    args = parse_args(sys.argv[1:])
    method = args.method
    client = SierraClient(args.url)
    client.toggle_progress(True)
    if method == 'fasta':
        fasta_handler(client, args)
    elif method == 'mutations':
        mutations_handler(client, args)
    elif method == 'patterns':
        patterns_handler(client, args)
    elif method == 'introspection':
        introspection_handler(client, args)
    else:
        print('Program error. If you keep seeing this message please submit \n'
              'an issue on https://github.com/hivdb/sierra-client/issues.',
              file=sys.stderr)
        exit(127)
    """



# -*- coding: utf-8 -*-
import json
import click

from .. import fragments
from .cli import cli


@cli.command()
@click.argument('mutations', nargs=-1, required=True)
@click.option('-q', '--query', type=click.File('r'),
              help=('A file contains GraphQL fragment definition '
                    'on `MutationsAnalysis`.'))
@click.option('-o', '--output', default='-', type=click.File('w'),
              help='File path to store the JSON result.')
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def mutations(ctx, mutations, query, output, ugly):
    """
    Run drug resistance and other analysis for PR, RT and/or IN mutations.
    For Example:

    \b
    sierrapy mutations PR:E35E_D RT:T67- IN:M50MI

    Use command "sierrapy patterns" instead if you want to run multiple sets
    of mutations in one request.
    """
    if query:
        query = query.read()
    else:
        query = fragments.MUTATIONS_ANALYSIS_DEFAULT
    result = ctx.obj['CLIENT'].mutations_analysis(mutations, query)
    json.dump(result, output, indent=None if ugly else 2)



# -*- coding: utf-8 -*-

from collections import OrderedDict

import requests
from tqdm import tqdm
from gql import gql, Client
from gql.transport.http import HTTPTransport
from graphql.execution import ExecutionResult
from graphql.language.printer import print_ast
from requests.exceptions import HTTPError

VERSION = '0.2.0'
DEFAULT_URL = 'https://hivdb.stanford.edu/graphql'


class ResponseError(Exception):
    pass


class RequestsHTTPTransportWOrderedDict(HTTPTransport):
    def __init__(self, url, timeout=None, **kwargs):
        super(RequestsHTTPTransportWOrderedDict, self).__init__(url, **kwargs)
        self.default_timeout = timeout

    def execute(self, document, variable_values=None, timeout=None):
        query_str = print_ast(document)
        payload = {
            'query': query_str,
            'variables': variable_values or {}
        }
        request = requests.post(
            self.url,
            json=payload,
            headers=self.headers,
            timeout=timeout or self.default_timeout)
        request.raise_for_status()
        result = request.json(object_pairs_hook=OrderedDict)
        assert 'errors' in result or 'data' in result, \
            'Received non-compatible response "{}"'.format(result)
        return ExecutionResult(
            errors=result.get('errors'),
            data=result.get('data')
        )


class SierraClient(object):

    def __init__(self, url=DEFAULT_URL):
        self.url = url
        self._client = None
        self._progress = False

    def toggle_progress(self, flag='auto'):
        if flag == 'auto':
            self._progress = not self._progress
        else:
            self._progress = bool(flag)

    @property
    def client(self):
        if self._client is None:
            transport = \
                RequestsHTTPTransportWOrderedDict(self.url, timeout=300)
            transport.headers = {
                'Content-Type': 'application/json',
                'User-Agent': 'sierra-client (python)/{}'.format(VERSION)
            }
            self._client = Client(
                transport=transport,
                fetch_schema_from_transport=True)
        return self._client

    def execute(self, document, variable_values=None):
        try:
            return self.client.execute(
                document, variable_values=variable_values)
        except HTTPError as e:
            errors = [
                e['exception']['detailMessage']
                if 'detailMessage' in e['exception']
                else 'Unknown server error'
                for e in e.response.json()['errors']
                if 'exception' in e]
            raise ResponseError(
                'Sierra GraphQL webservice returned errors:\n - ' +
                '\n - '.join(errors))

    def get_introspection(self):
        return self.client.introspection

    def _sequence_analysis(self, sequences, query):
        result = self.execute(
            gql("""
                query sierrapy($sequences:[UnalignedSequenceInput]!) {{
                    viewer {{
                        sequenceAnalysis(sequences:$sequences) {{
                            ...F0
                        }}
                    }}
                }}
                fragment F0 on SequenceAnalysis {{
                    {query}
                }}
                """.format(query=query)),
            variable_values={"sequences": sequences})
        return result['viewer']['sequenceAnalysis']

    def _pattern_analysis(self, patterns, query, **kw):
        enable_hivalg = 'algorithms' in kw or 'customAlgorithms' in kw
        extraparams = ''
        if enable_hivalg:
            extraparams = (', $algorithms:[ASIAlgorithm], '
                           '$customAlgorithms: [CustomASIAlgorithm]')
        variables = {"patterns": patterns}
        if enable_hivalg:
            variables['algorithms'] = kw.get('algorithms')
            variables['customAlgorithms'] = kw.get('custom_algorithms')
        result = self.execute(
            gql("""
                query sierrapy($patterns:[[String]!]!{extraparams}) {{
                    viewer {{
                        patternAnalysis(patterns:$patterns) {{
                            ...F0
                        }}
                    }}
                }}
                fragment F0 on MutationsAnalysis {{
                    {query}
                }}
                """.format(query=query, extraparams=extraparams)),
            variable_values=variables)
        return result['viewer']['patternAnalysis']

    def iter_sequence_analysis(self, sequences, query, step=20):
        if self._progress:
            pbar = tqdm(total=len(sequences))
        while sequences:
            for result in self._sequence_analysis(sequences[:step], query):
                yield result
            sequences = sequences[step:]
            self._progress and pbar.update(step)

    def iter_pattern_analysis(self, patterns, query, step=20, **kw):
        if self._progress:
            pbar = tqdm(total=len(patterns))
        while patterns:
            for result in self._pattern_analysis(patterns[:step], query, **kw):
                yield result
            patterns = patterns[step:]
            self._progress and pbar.update(step)

    def sequence_analysis(self, sequences, query, step=20):
        return list(self.iter_sequence_analysis(sequences, query, step))

    def pattern_analysis(self, patterns, query, step=20, **kw):
        return list(self.iter_pattern_analysis(patterns, query, step, **kw))

    def mutations_analysis(self, mutations, query):
        result = self.execute(
            gql("""
                query sierrapy($mutations:[String]!) {{
                    viewer {{
                        mutationsAnalysis(mutations:$mutations) {{
                            ...F0
                        }}
                    }}
                }}
                fragment F0 on MutationsAnalysis {{
                    {query}
                }}
                """.format(query=query)),
            variable_values={"mutations": mutations})
        return result['viewer']['mutationsAnalysis']

    def current_version(self):
        result = self.execute(
            gql("""
                query sierrapy {
                    viewer {
                        currentVersion { text, publishDate }
                    }
                }
                """))
        return result['viewer']['currentVersion']



# -*- coding: utf-8 -*-

import json
import click
from collections import OrderedDict
from voluptuous import Schema, Required, MultipleInvalid, ALLOW_EXTRA

GENES = OrderedDict([
    ('PR', 99),
    ('RT', 560),
    ('IN', 288)
])


schema = Schema([{
    Required('inputSequence'): {
        Required('header'): str
    },
    Required('alignedGeneSequences'): [{
        Required('gene'): {
            Required('name'): str
        },
        Required('firstAA'): int,
        Required('lastAA'): int,
        Required('alignedNAs'): str,
        Required('prettyPairwise'): {
            Required('positionLine'): [str],
            Required('alignedNAsLine'): [str]
        }
    }]
}], extra=ALLOW_EXTRA)


@click.option('--gap-handling', default="hxb2strip",
              type=click.Choice(['squeeze', 'hxb2strip']),
              help=('Specify how you want the recipe to handle the gaps.\n\n'
                    'Specify "squeeze" to keep every gap in the result '
                    'alignment; "hxb2_strip" to strip out non-HXB2 columns.'))
@click.pass_context
def alignment(ctx, gap_handling):
    """Export aligned pol sequences from Sierra result."""
    output = ctx.obj['OUTPUT']
    sequences = json.load(ctx.obj['INPUT'])
    try:
        schema(sequences)
    except MultipleInvalid as e:
        raise click.ClickException(str(e))
    for seq in sequences:
        seqheader = seq['inputSequence']['header']
        concat_seqs = ''
        geneseqs = {gs['gene']['name']: gs
                    for gs in seq['alignedGeneSequences']}
        for gene, genesize in GENES.items():
            geneseq = geneseqs.get(gene)
            if geneseq:
                first_aa = geneseq['firstAA']
                last_aa = geneseq['lastAA']
                naseq = geneseq['alignedNAs']
            else:
                first_aa = 1
                last_aa = genesize
                naseq = '.' * genesize * 3
            if gap_handling == 'hxb2strip':
                naseq = (('.' * (first_aa - 1) * 3) +
                         naseq +
                         '.' * (genesize - last_aa) * 3)
            else:  # gap_handling == 'squeeze'
                raise NotImplementedError()
            concat_seqs += naseq
        output.write('>{}\n{}\n'.format(seqheader, concat_seqs))



# -*- coding: utf-8 -*-
import click

from .cli import cli
from .. import recipes


@cli.group()
@click.option('--input', default='-', type=click.File('r'),
              help='JSON result from Sierra web service.')
@click.option('--output', default='-', type=click.File('w'),
              help='File path to store the result.')
@click.pass_context
def recipe(ctx, input, output):
    """Post process Sierra web service output."""
    ctx.obj['INPUT'] = input
    ctx.obj['OUTPUT'] = output


for subcommand in recipes.__all__:
    subcommand = getattr(recipes, subcommand)
    recipe.command()(subcommand)



# -*- coding: utf-8 -*-
import click
import json
from itertools import chain

from .. import fastareader, fragments
from .cli import cli


@cli.command()
@click.argument('fasta', nargs=-1, type=click.File('r'), required=True)
@click.option('-q', '--query', type=click.File('r'),
              help=('A file contains GraphQL fragment definition '
                    'on `SequenceAnalysis`.'))
@click.option('-o', '--output', default='-', type=click.File('w'),
              help='File path to store the JSON result.')
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def fasta(ctx, fasta, query, output, ugly):
    """
    Run alignment, drug resistance and other analysis for one or more
    FASTA-format files contained HIV-1 pol DNA sequences.
    """
    sequences = list(chain(*[fastareader.load(fp) for fp in fasta]))
    if query:
        query = query.read()
    else:
        query = fragments.SEQUENCE_ANALYSIS_DEFAULT
    result = ctx.obj['CLIENT'].sequence_analysis(sequences, query)
    json.dump(result, output, indent=None if ugly else 2)



# -*- coding: utf-8 -*-

import os

_globals = globals()
__all__ = []

dirpath = os.path.dirname(__file__)

for path in os.listdir(dirpath):
    if path.endswith('.gql'):
        name, _ = os.path.splitext(path)
        with open(os.path.join(dirpath, path)) as fp:
            _globals[name.upper()] = fp.read()
            __all__.append(name.upper())



# -*- coding: utf-8 -*-


def load(fp):
    sequences = []
    header = None
    curseq = ''
    for line in fp:
        if line.startswith('>'):
            if header and curseq:
                sequences.append({
                    'header': header,
                    'sequence': curseq
                })
            header = line[1:].strip()
            curseq = ''
        elif line.startswith('#'):
            continue
        else:
            curseq += line.strip()
    if header and curseq:
        sequences.append({
            'header': header,
            'sequence': curseq
        })
    return sequences



# -*- coding: utf-8 -*-
import json
import click

from .cli import cli


@cli.command()
@click.option('-o', '--output', default='-', type=click.File('w'),
              help='File path to store the JSON result.')
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def introspection(ctx, output, ugly):
    """Output introspection of Sierra GraphQL web service."""
    result = ctx.obj['CLIENT'].get_introspection()
    json.dump(result, output, indent=None if ugly else 2)



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import re
import ast
import setuptools

_version_re = re.compile(r'VERSION\s+=\s+(.*)')

with open('sierrapy/sierraclient.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


def strip_comments(l):
    return l.split('#', 1)[0].strip()


def req(filename):
    with open(os.path.join(os.getcwd(), filename)) as fp:
        requires = set([strip_comments(l) for l in fp.readlines()])
        requires -= set([''])
    return list(requires)


setup_params = dict(
    name="sierrapy",
    version=version,
    url="https://github.com/hivdb/sierra-client/tree/master/python",
    author='Philip Tzou',
    author_email="philiptz@stanford.edu",
    description='A Client of HIVdb Sierra GraphQL Webservice.',
    packages=['sierrapy', 'sierrapy/fragments',
              'sierrapy/recipes', 'sierrapy/commands'],
    install_requires=req('requirements.txt'),
    # tests_require=reqs('test-requirements.txt'),
    include_package_data=True,
    entry_points={'console_scripts': [
        'sierrapy = sierrapy.cmds:main',
    ]},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics'],
    # test_suite="nose.collector",
    zip_safe=True)

if __name__ == '__main__':
    setuptools.setup(**setup_params)



# -*- coding: utf-8 -*-
import re
import json
import click
from sierrapy import fragments

from .cli import cli


@cli.command()
@click.argument('patterns', nargs=-1, required=True, type=click.File('r'))
@click.option('-q', '--query', type=click.File('r'),
              help=('A file contains GraphQL fragment definition '
                    'on `MutationsAnalysis`.'))
@click.option('-o', '--output', default='-', type=click.File('w'),
              help='File path to store the JSON result.')
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def patterns(ctx, patterns, query, output, ugly):
    """
    Run drug resistance and other analysis for one or more files contains
    lines of PR, RT and/or IN mutations based on HIV-1 type B consensus.
    Each line is treated as a unique pattern. For example:

    \b
    RT:M41L + RT:M184V + RT:L210W + RT:T215Y
    PR:L24I + PR:M46L + PR:I54V + PR:V82A

    The following delimiters are supported: commas (,), plus signs (+),
    semicolon(;), whitespaces and tabs. The consensus sequences can be
    retrieved from HIVDB website: <https://goo.gl/ZBthkt>.
    """
    ptns = []
    for fp in patterns:
        for line in fp:
            ptns.append(re.split(r'[,;+ \t]+', line.strip()))
    if query:
        query = query.read()
    else:
        query = fragments.MUTATIONS_ANALYSIS_DEFAULT
    result = ctx.obj['CLIENT'].pattern_analysis(patterns, query)
    json.dump(result, output, indent=None if ugly else 2)



# -*- coding: utf-8 -*-

QUERY_SEQUENCE_ANALYSIS = """
    inputSequence { header },
    subtypeText,
    validationResults {
        level,
        message
    },
    alignedGeneSequences {
        firstAA, lastAA,
        gene { name },
        mutations {
            consensus,
            position,
            AAs,
            isInsertion,
            isDeletion,
            isApobecDRM
        },
        SDRMs:mutations(filterOptions:[SDRM]) {
            text
        },
        prettyPairwise {
            positionLine,
            refAALine,
            alignedNAsLine,
            mutationLine
        }
    },
    drugResistance {
        gene { name },
        drugScores {
            drugClass { name },
            drug {
                name,
                displayAbbr,
            },
            score,
            partialScores {
                mutations {
                    text,
                    primaryType,
                    comments { type, text }
                },
                score
            },
            text
        }
    }
"""
QUERY_MUTATIONS_ANALYSIS = """
    validationResults {
        level,
        message
    },
    drugResistance {
        gene { name },
        drugScores {
            drugClass { name },
            drug {
                name,
                displayAbbr,
            },
            score,
            partialScores {
                mutations {
                    consensus,
                    position,
                    AAs,
                    text,
                    primaryType,
                    comments { type, text }
                },
                score
            },
            text
        }
    }
"""
